module phobos.gc;

private extern (C) void free (void *data);
private extern (C) void *malloc (int size);
private extern (C) void memset (void *data, int value, int size);
private extern (C) void memmove (void *dest, void *src, int length);

struct gc_block
{
    gc_type *type; /* Link to the block type */
    int length; /* Size of the allocated block in bytes */
    
    gc_block *left; /* Left node, can be NULL */
    gc_block *right; /* Right node, can be NULL */
    gc_block *parent; /* Link to the parent */
    char color; /* Node color; 'r' (red) or 'b' (black) */
    bit immortal; /* Cannot die */
    bit marked; /* Set if it has been accessed */
    void *data; /* Start of the data, normally this address plus four */

    /* Depth of this block from the root */
    int depth ()
    {
        gc_block *node = this;
        int c;

        for (c = 0; node != gc_block_node; c ++, node = node.parent)
            { }
        return c;
    }

    /* Mark the block */
    void mark (void *data)
    {
        type.mark (this, data);
    }

    /* Insert this block into the list */
    void insert ()
    {
        gc_block *o_left = left;
        gc_block *o_right = right;
        gc_block *gimp;
        gc_block *node;

        left = right = null;
        __insert ();
        color = 'r';
        node = this;

        while (node != gc_block_node && node.parent.color == 'r')
        {
            if (node.parent == node.parent.parent.left)
            {
                gimp = node.parent.parent.right;
                if (gimp && gimp.color == 'r')
                {
                    node.parent.color = 'b';
                    gimp.color = 'b';
                    node.parent.parent.color = 'r';
                    node = node.parent.parent;
                }
                else
                {
                    if (node == node.parent.right)
                    {
                        node = node.parent;
                        node.left_rotate ();
                    }

                    node.parent.color = 'b';
                    node.parent.parent.color = 'r';
                    node.parent.parent.right_rotate ();
                }
            }
            else
            {
                gimp = node.parent.parent.left;
                if (gimp && gimp.color == 'r')
                {
                    node.parent.color = 'b';
                    gimp.color = 'b';
                    node.parent.parent.color = 'r';
                    node = node.parent.parent;
                }
                else
                {
                    if (node == node.parent.left)
                    {
                        node = node.parent;
                        node.right_rotate ();
                    }

                    node.parent.color = 'b';
                    node.parent.parent.color = 'r';
                    node.parent.parent.left_rotate ();
                }
            }
        }

        gc_block_node.color = 'b';

        if (o_left)
        {
            o_left.parent = null;
            o_left.insert ();
        }
        
        if (o_right)
        {
            o_right.parent = null;
            o_right.insert ();
        }
    }

    int compare (gc_block *other)
    {
        if ((void *) this > (void *) other)
            return -1;
        return 1;
    }

    /* Return whether this block contains the data */
    bit contains (void *data)
    {
        return data >= (void *) this && data < (void *) this + gc_block.size + length;
    }

    void left_rotate ()
    {
        gc_block *gimp;

        gimp = right;
        right = gimp.left;
        if (gimp.left)
            gimp.left.parent = this;
        gimp.parent = parent;

        if (!parent)
            gc_block_node = gimp;
        else if (this == parent.left)
            parent.left = gimp;
        else
            parent.right = gimp;
            
        gimp.left = this;
        parent = gimp;
    }

    void right_rotate ()
    {
        gc_block *gimp;
        
        gimp = left;
        left = gimp.right;
        if (gimp.right)
            gimp.right.parent = this;
        gimp.parent = parent;
        if (!parent)
            gc_block_node = gimp;
        else if (this == parent.right)
            parent.right = gimp;
        else
            parent.left = gimp;

        gimp.right = this;
        parent = gimp;
    }

    /* Remove the node from the tree */

    void remove ()
    {
        if (gc_block_node == this)
            gc_block_node = null;
            
        if (!parent)
            { }
        else if (parent.left == this)
            parent.left = null;
        else
            parent.right = null;

        if (left)
            left.insert ();
        if (right)
            right.insert ();
    }

    private void __insert ()
    {
        gc_block *base = gc_block_node;
        
        if (!gc_block_node)
        {
            gc_block_node = this;
            parent = null;
            return;
        }

        while (1)
        {
            if (compare (base) > 0)
            {
                if (!base.left)
                {
                    base.left = this;
                    parent = base;
                    return;
                }

                base = base.left;
            }
            else
            {
                if (!base.right)
                {
                    base.right = this;
                    parent = base;
                    return;
                }

                base = base.right;
            }
        }
    }
}

struct gc_type
{
    void (*mark) (gc_block *block, void *point); /* Called when this block is marked. */
    void (*done) (gc_block *block); /* Called when this object may need deletion. */
                                    /* In any case, clear block.marked to false */
}

protected gc_block *gc_block_node = null; /* Root node */
protected void *gc_stack_top = null;
int gc_usage = 0; /* Current memory usage */
int gc_peak_usage = 0; /* Peak memory usage */

void gc_black_mark (gc_block *block, void *point)
{
    gc_mark_memory (block.data, block.data + block.length);
}

void gc_black_done (gc_block *block)
{
    if (block.marked)
        block.marked = false;
    else
        gc_free (block);
}

gc_type gc_black_type = { null, null };//&gc_black_mark, &gc_black_done };

/* This should be called by types to mark the data and continue its processing. */
void gc_mark (void *data)
{
    gc_block *block;
    gc_type *type;
    
    block = gc_find (data);
    if (!block || block.marked)
        return;
    block.marked = true;
    block.type.mark (block, data);
}

void gc_mark (gc_block *block)
{
    if (!block || block.marked)
        return;
    block.marked = true;
    block.type.mark (block, block.data);        
}

/* Try to find the block for this data, or return NULL if it couldn't be found. */
gc_block *gc_find (void *data)
{
    gc_block *node = gc_block_node;
   
    while (node)
    {
        if (node.contains (data))
            return node;
        if ((void *) node > data)
            node = node.left;
        else
            node = node.right;
    }

    return node;
}

/* Force the freeing of a block - doesn't call the del method */
void gc_free (gc_block *block)
{
    gc_usage -= block.length;
    block.remove ();
    free (block);
}

/* Allocate and clear alien data (not managed) */
void *gc_alloc_alien (int size)
{
    void *data;

    data = malloc (size);
    memset (data, 0, size);
    return data;
}

/* Core of the various allocators */
protected gc_block *gc_alloc_core (int size, gc_type *type)
{
    gc_block *block;

    block = (gc_block *) malloc (size + gc_block.size);
    memset (block, 0, gc_block.size);

    block.type = type;
    block.length = size;
    block.data = (void *) block + gc_block.size;
    block.insert ();

    gc_usage += block.length;
    if (gc_usage > gc_peak_usage)
        gc_peak_usage = gc_usage;

    return block;
}

/* Allocate a block and clear it */
gc_block *gc_alloc (int size, gc_type *type)
{
    gc_block *block;

    block = gc_alloc_core (size, type);
    memset (block.data, 0, size);

    return block;
}

/* Allocate a duplicate of memory */
gc_block *gc_alloc_dup (void *data, int size, gc_type *type)
{
    gc_block *block;

    block = gc_alloc_core (size, type);
    memmove (block.data, data, size);

    return block;
}

void gc_dump (gc_block *block)
{
    if (block)
    {
        gc_dump (block.left);
        printf ("%p - %d byte%s, %d\n", block.data, block.length, block.length == 1 ? (char *) "" : (char *) "s", block.depth ());
        gc_dump (block.right);
    }
}

void gc_dump ()
{
    gc_dump (gc_block_node);
}

void gc_init ()
{
    int top;

    gc_stack_top = (void *) &top + 32;
    gc_black_type.mark = &gc_black_mark;
    gc_black_type.done = &gc_black_done;
}

/* Mark any immortal blocks */

void gc_mark_immortal (gc_block *block)
{
    if (block)
    {
        gc_mark_immortal (block.left);
        gc_mark_immortal (block.right);
        if (block.immortal)
            gc_mark (block);
    }
}

void gc_mark_immortal () { gc_mark_immortal (gc_block_node); }

/* Scan a block of memory and mark as necessary */

void gc_mark_memory (void *start, void *end)
{
    while ((int) start + 4 <= (int) end)
    {
        void *data = *(void **) start;

        gc_mark (*(void **) start);
        start += 4;
    }
}

/* Free any blocks left unmarked */

void gc_done (gc_block *block)
{
    if (block)
    {
        gc_done (block.left);
        gc_done (block.right);
        block.type.done (block);
    }
}

void gc_done ()
{
    gc_done (gc_block_node);
}

/* Do a full collect */

void gc_full_collect ()
{
    int bottom;
    
    gc_mark_immortal ();
    gc_mark_memory (&bottom, gc_stack_top);
    gc_done ();
}

/* Add an immortal block */

void gc_add (gc_block *node)
{
    node.insert ();
}

extern (C)
void *_d_new (int count, int size)
{
    void *data;
    
    if (count < 0)
        throw new Error ("tried to allocate negative number of object");
    if (size < 0)
        throw new Error ("system failure: allocation size below zero");
    data = gc_alloc (count * size, &gc_black_type).data;
    return data;
}

extern (C)
void *_d_dup (void *data, int size)
{
    if (size < 0)
        throw new Error ("system failure: allocation size below zero");
    return gc_alloc_dup (data, size, &gc_black_type).data;
}

