module object;

extern (C) static int printf(char *, ...);

/* This contains all the information about a format conversion. */

struct FormatSpecifier
{
    bit alternate; /* Use alternate form ("#"). */
    bit zero; /* Pad with zeros instead of spaces ("0"). */
    bit left; /* Left-adjust and pad on the right instead ("-"). */
    bit blank; /* Put a blank before a positive number or empty string (" "). */
    bit plus; /* Always place a sign before a number ("+"). */
    bit locale; /* Use locale printing and thousands separators ("'"). */
    bit altdigits; /* Use the locale's alternate output digits, if any ("I"). */
    bit capital; /* Capitalise hex digits ("^"). */
    bit precapital; /* Capitalise the "0x" prefix ("^" before "#"). */
    
    int width = -1; /* Field width, negative if unspecified. */
    int precision = -1; /* Field precision, negative if unspecified. */

    char code; /* Actual specifier */
    
    /* Reset flags that trump one another */
    void clear ()
    {
        if (left) zero = false;
        if (plus) blank = false;
    }
    
    int radix () { return code == "b" ? 2 : code == "o" ? 8 : code == "x" ? 16 : 10; }
    bit isReal () { return code == "e" || code == "f" || code == "g"; }
}

class Object
{
    char[] toString () { return "Object"; }
    uint toHash () { return (uint) (void *) this; }
    int cmp (Object o) { return (int) (void *) this - (int) (void *) o; }
    int eq (Object o) { return this === o; }
    char[] toFormat (FormatSpecifier *spec) { return toString (); }
}

struct Interface
{
    ClassInfo classinfo;
    void *[] vtbl;
    int offset;         // offset to Interface 'this' from Object 'this'
}

class SymbolInfo : Object
{
    char [] name; /* Name of the symbol */
    SymbolInfo parent; /* Parent of this symbol (Such as a ModuleInfo) */

    /* Search for the module this symbol is in or null */
    ModuleInfo findModule ()
    {
        SymbolInfo symbol = this;

        while (symbol)
        {
            if (cast (ModuleInfo) symbol)
                return cast (ModuleInfo) symbol;
            symbol = symbol.parent;
        }

        return null;
    }

    char [] mangle ()
    {
        char [] text;

        for (SymbolInfo sym = this; sym; sym = sym.parent)
            text = itoa (sym.name.length) ~ sym.name ~ "_" ~ text;

        text = "_D" ~ text;
        
        if (getType ())
            return text ~ "_" ~ getType ().mangle ();
        return text;
    }

    TypeInfo getType ()
    {
        return null;
    }
}

/* Type description.  You can get the TypeInfo of any value by retrieving the
 * ".typeinfo" property of it; for example:
 *
 *     (4).typeinfo
 */

class TypeInfo
{
    /* Nested type (such as for the type of a pointer) or null for none */
    TypeInfo next;
    
    /* Get a hash value that should be fairly unique for this value. */
    uint getHash (void *p) { return (uint) p; }

    /* Return zero if these objects are inequal, anything else if they are. */
    int equals(void *p1, void *p2) { return p1 == p2; }

    /* Compare the objects and return negative when *p1 < *p2,
                                      positive when *p1 > *p2,
                                      zero when *p1 == *pw */
                                      
    int compare (void *p1, void *p2) { return 0; }

    /* Get the size of an instance in bytes. */
    int tsize () { return 0; }

    /* Swap the values so that the contents of *p1 become *p2 and vice versa. */
    
    void swap (void *p1, void *p2)
    {
        int i;
        int n = tsize();
        for (i = 0; i < n; i++)
        {
            byte t;

            t = ((byte *) p1) [i];
            ((byte *) p1) [i] = ((byte *) p2) [i];
            ((byte *) p2) [i] = t;
        }
    }

    byte [] tinit () { throw new Error (kind () ~ " doesn't implement tinit"); byte [] array; return array; }
    char [] kind () { return "unknown"; }

    /* Returns true if this is float, double, extended, complex, or imaginary */
    bit isReal () { return false; }

    bit isBasic () { return false; } /* Returns true if this is a basic type - bit, float, int, etc. */
    
    int toInt (void *a) { throw new Error ("cannot cast type to int"); return 0; }
    bit toBit (void *a) { throw new Error ("cannot cast type to bit"); return false; }

    void print () { printf ("%.*s\n", toString ()); }

    char [] toString () { return toString (""); }
    
    char [] toString (char [] name)
    {
        char [] text;
        
        if (next)
            text = kind () ~ " " ~ next.toString ();
        else
            text = kind ();

        if (name)
            text ~= " " ~ name;
        return text;
    }

    /* Base character to mangle on */
    char [] mangleChar () { throw new Error ("mangleChar must be implemented"); return "foo"; }

    /* Get the mangling for this type */
    char [] mangle () { return mangleChar (); }

    int eq (Object other)
    {
        if (cast (TypeInfo) other)
        {
            TypeInfo o = cast (TypeInfo) other;

            if (o.kind () != kind ())
                return false;
            return true;
        }
        else
            return false;
    }
    
    /* Format the value as appropriate */
    char [] format (FormatSpecifier *spec, void *data) { throw new Error ("format must be implemented for " ~ this.classinfo.name); return "foo"; }
}

/* Class description.  There is one of these per class; you can get the
 * ClassInfo of an instance by getting the ".classinfo" property of it;
 * for example:
 *
 *     foo.classinfo
 */

extern (C) void *_d_dup (void *data, int size);

class ClassInfo : SymbolInfo
{
    byte [] init; /* Initial content before calling a constructor */
    void *[] vtbl; /* virtual function pointer table */
        /* A pointer to the vtbl is the first field of an instance.  The first
         * field of the vtbl is to the ClassInfo; this is followed by pointers
         * to methods.
         */
         
    Interface [] interfaces; /* List of interfaces applied to this class */
    ClassInfo base; /* Base class pointer */
    void *destructor; /* Pointer to the destructor */
    void (*_invariant)(Object);
    uint flags;
    //  1:          // IUnknown

    /* These don't include base class values */
    FieldInfo [] fields; /* List of fields in this class */
    MethodInfo [] methods; /* List of methods in the class */
    
    /* Create a new instance of this class, not calling any constructors */
    Object create ()
    {
        return (Object) _d_dup (init, init.length);
    }
    
    int eq (Object other)
    {
        ClassInfo info = cast (ClassInfo) other;
        
        if (!info)
            return false;
        if (info.name != name)
            return false;
        if (base && !info.base)
            return false;
        if (base && !base.eq (info.base))
            return false;
        if (parent && !info.parent)
            return false;
        if (parent && !parent.eq (info.parent))
            return false;
        return true;
    }
}

class StructInfo : SymbolInfo
{
    byte [] init; /* Initial content */
    
    FieldInfo [] fields; /* List of fields in this struct */
    MethodInfo [] methods; /* List of methods in this struct */
    
    int eq (Object other)
    {
        StructInfo info = cast (StructInfo) other;
        
        if (!info)
            return false;
        if (info.name != name)
            return false;
        if (parent && !info.parent)
            return false;
        if (parent && !parent.eq (info.parent))
            return false;
        return true;
    }
}

enum Protection
{
    Undefined,
    None,
    Private,
    Protected,
    Public,
    Export,
}

class FieldInfo : SymbolInfo
{
    TypeInfo type; /* Type of the field */
    int offset; /* Data offset from the start of the class/struct to this field */

    Protection protection; /* Protection level */

    bit isPublic () { return protection == Protection.Public; }
    bit isPrivate () { return protection == Protection.Private; }
    bit isProtected () { return protection == Protection.Protected; }
    TypeInfo getType () { return type; }
}

import std.type;
import std.format;

class MethodInfo : SymbolInfo
{
    TypeFunction type; /* Function descriptor */
    int offset; /* Offset to this method in the vtable */
    void *pointer; /* Pointer to the method */

    Protection protection; /* Protection level */

    bit isPublic () { return protection == Protection.Public; }
    bit isPrivate () { return protection == Protection.Private; }
    bit isProtected () { return protection == Protection.Protected; }
    TypeInfo getType () { return type; }

    char [] toString ()
    {
        if (name == "_ctor")
            return type.toString ("this");
        return type.toString (name);
    }
}

/* Module Description */

class ModuleInfo : SymbolInfo
{
    SymbolInfo [] members; /* Symbols contained by this module */
    void (*ctor) (); /* Static constructor, calls all static constructors */

    static ModuleInfo [] loaded; /* List of loaded modules */

    /* Search for the module with this name, returning it or null */
    static ModuleInfo findModule (char [] name)
    {
        for (int c = 0; c < loaded.length; c ++)
            if (loaded [c].name == name)
                return loaded [c];
                
        return null;
    }

    /* Search for a symbol with this name, returning it or null */
    SymbolInfo findSymbol (char [] name)
    {
        for (int c = 0; c < members.length; c ++)
            if (members [c].name == name)
                return members [c];

        return null;
    }
}

/* Enumeration Description */

class EnumInfo : SymbolInfo
{
    TypeInfo memtype; /* Type of the members */
    long max; /* Maximum value */
    long min; /* Minimum value */
    long def; /* Default value */

    struct Value
    {
        long value;
        char [] name;
    }

    Value [] values; /* List of values */

    /* Return whether this value is valid in the enumeration */
    bit hasValue (long val)
    {
        for (int c = 0; c < values.length; c ++)
            if (values [c].value == val)
                return true;
        return false;
    }

    /* Get the name of the value - null if not found */
    char [] valueName (long val)
    {
        for (int c = 0; c < values.length; c ++)
            if (values [c].value == val)
                return values [c].name;
        return null;
    }
}

class Exception : Object
{
    char[] msg;

    this(char[] msg)
    {
    this.msg = msg;
    }

    void print()
    {
    printf("%.*s\n", msg);
    }

    char[] toString () { return msg; }
}

class Error : Exception
{
    Error next;

    this (char [] msg)
    {
    super (msg);
    }

    this (char [] msg, Error next)
    {
    super (msg);
    this.next = next;
    }
}

class DivideByZeroError : Error
{
    this (char [] msg)
    {
        super (msg);
    }
}

struct generic
{
    TypeInfo type;
    void *data;
    
    uint getHash () { return type.getHash (data); }
    int equals (void *p2) { return type.equals (data, p2); }
    int compare (void *p2) { return type.compare (data, p2); }
    int tsize () { return type.tsize (); }
    void swap (void *p2) { type.swap (data, p2); }
    byte [] tinit () { return type.tinit (); }
    char [] kind () { return type.kind (); }
    bit isReal () { return type.isReal (); }
    bit isBasic () { return type.isBasic (); }
    int toInt () { return type.toInt (data); }
    bit toBit () { return type.toBit (data); }
    char [] format (FormatSpecifier *spec) { return type.format (spec, data); }
}

generic genericFor (TypeInfo type, void *data)
{
    int size = type.tsize ();
    byte [] array = new byte [size];
    generic g;
    
    g.type = type;
    g.data = (void *) array;
    array [] = ((byte *) data) [0 .. size];
    return g;
}
