
// Copyright (c) 2001-2002 by Digital Mars
// All Rights Reserved
// www.digitalmars.com

import std.c.stdio;
import std.string;

/***********************************
 */

class FileError : Error
{
    import std.syserror;

    uint errno;         // operating system error code

    this(char[] name)
    {
    this(name, "file I/O");
    }

    this(char[] name, char[] message)
    {
    msg = name ~ ": " ~ message;
    }

    this(char[] name, uint errno)
    {
        char *message = strerror (errno);
        
    this(name, message [0 .. strlen (message)]);
    this.errno = errno;
    }
}

/***********************************
 * Basic File operations.
 */

version (Win32)
{
    import windows;
}

/********************************************
 * Read a file.
 * Returns:
 *  array of bytes read
 */

version (Win32)
{
    byte[] read(char[] name)
    {
        DWORD size;
        DWORD numread;
        HANDLE h;
        byte[] buf;
        char* namez;

        namez = toStringz(name);
        h = CreateFileA(namez,GENERIC_READ,FILE_SHARE_READ,null,OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,(HANDLE)null);
        if (h == INVALID_HANDLE_VALUE)
            goto err1;

        size = GetFileSize(h, null);
        buf = new byte[size];
    
        if (ReadFile(h,buf,size,&numread,null) != 1)
        goto err2;

        if (numread != size)
        goto err2;

        if (!CloseHandle(h))
            goto err;

        return buf;

    err2:
        CloseHandle(h);
    err:
        delete buf;
    err1:
        throw new FileError(name, GetLastError());
    }
}
else version (Linux)
{
    byte [] read (char [] name)
    {
        byte [] buf;
        FILE *file;
        int size;

        file = fopen ((char *) name, "rb");
        if (file == null)
            goto err;

        if (fseek (file, 0, SEEK_END) < 0)
            goto err2;
            
        if ((size = ftell (file)) < 0)
            goto err2;

        if (fseek (file, 0, SEEK_SET) < 0)
            goto err2;
            
        buf = new byte [size];
        
        if (fread (buf, 1, size, file) != size)
            goto err2;
            
        fclose (file);
        return buf;

    err2:
        fclose (file);
    err:
        throw new FileError (name, "Error reading file");
    }
}

/*********************************************
 * Write a file.
 * Returns:
 *  0   success
 */

version (Win32)
{
    void write(char[] name, byte[] buffer)
    {
        HANDLE h;
        DWORD numwritten;

        h = CreateFileA(toStringz(name),GENERIC_WRITE,0,null,CREATE_ALWAYS,
        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,(HANDLE)null);
        if (h == INVALID_HANDLE_VALUE)
        goto err;

        if (WriteFile(h,buffer,buffer.length,&numwritten,null) != 1)
            goto err2;

        if (buffer.length != numwritten)
            goto err2;
    
        if (!CloseHandle(h))
            goto err;
        return;

    err2:
        CloseHandle(h);
    err:
        throw new FileError(name, GetLastError());
    }
}
else version (Linux)
{
    void write (char [] name, byte [] buffer)
    {
        FILE *file;

        file = fopen ((char *) name, "wb");
        if (file == null)
            goto err;
        if (fwrite (buffer, 1, buffer.length, file) != buffer.length)
            goto err2;
        fclose (file);
        return;
        
    err2:
        fclose (file);
    err:
        throw new FileError (name, "Couldn't write file");
    }
}

/*********************************************
 * Append to a file.
 */

version (Win32)
{
    void append(char[] name, byte[] buffer)
    {
        HANDLE h;
        DWORD numwritten;

        h = CreateFileA(toStringz(name),GENERIC_WRITE,0,null,OPEN_ALWAYS,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,(HANDLE)null);
        if (h == INVALID_HANDLE_VALUE)
            goto err;

        SetFilePointer(h, 0, null, FILE_END);
    
        if (WriteFile(h,buffer,buffer.length,&numwritten,null) != 1)
            goto err2;
    
        if (buffer.length != numwritten)
            goto err2;
    
        if (!CloseHandle(h))
            goto err;
        return;

    err2:
        CloseHandle(h);
    err:
        throw new FileError(name, GetLastError());
    }
}
else version (Linux)
{
    void append (char [] name, byte [] buffer)
    {
        FILE *file;

        file = fopen (name, "ab");
        if (!file)
            goto err;

        if (fwrite (buffer, 1, buffer.length, file) != buffer.length)
            goto err;
        fclose (file);
        return;

    err:
        if (file)
            fclose (file);
        throw new FileError (name, "Problem while appending");
    }
}


/***************************************************
 * Rename a file.
 */

version (Win32)
{
    void rename(char[] from, char[] to)
    {
        BOOL result;

        result = MoveFileA(toStringz(from), toStringz(to));
        if (!result)
            throw new FileError(to, GetLastError());
    }
}
else version (Linux)
{
    void rename (char [] from, char [] to)
    {
        std.c.stdio.rename (from, to);
    }
}

/***************************************************
 * Delete a file.
 */

version (Win32)
{
    void remove(char[] name)
    {
        BOOL result;

        result = DeleteFileA(toStringz(name));
        if (!result)
        throw new FileError(name, GetLastError());
    }
}
else version (Linux)
{
    void remove (char [] name)
    {
        std.c.stdio.remove (name);
    }
}

/***************************************************
 * Get file size.
 */

version (Win32)
{
    uint getSize(char[] name)
    {
        WIN32_FIND_DATA filefindbuf;
        HANDLE findhndl;
    
        findhndl = FindFirstFileA(toStringz(name), &filefindbuf);
        if (findhndl == (HANDLE)-1)
        {
            throw new FileError(name, GetLastError());
        }
        FindClose(findhndl);
        return filefindbuf.nFileSizeLow;
    }
}
else version (Linux)
{
    uint getSize (char [] name)
    {
        FILE *file;
        uint size;

        file = fopen (name, "rb");
        if (!file)
            goto err;

        if (fseek (file, 0, SEEK_END) < 0)
            goto err;
            
        size = ftell (file);
        fclose (file);
        return size;

    err:
        if (file) fclose (file);
        throw new FileError (name, "Error getting file size");
    }
}

/***************************************************
 * Get file attributes.
 */

version (Win32)
{
    uint getAttributes(char[] name)
    {
        uint result;

        result = GetFileAttributesA(toStringz(name));
        if (result == 0xFFFFFFFF)
        {
            throw new FileError(name, GetLastError());
        }
        return result;
    }
}

