/*

String formatting functions.  These are similar to printf, but typesafe; you
can use "%s" on any type and expect a fairly good result.

Ordinary characters are copied to the output.  "%%" is copied as a single "%".
Otherwise "%" is followed by zero or more flags, an optional minimum field
width, an optional precision, then finally the type specifier.

When an argument is going to be consumed, you can indicate a different place by
putting the new argument index, starting from one, followed by "$".  The
counting then starts from that place.  For example:

    fmt ("%2$d %d %1$d", 1, 2, 3);
    
Returns "2 3 1".  For the length specifiers that consume arguments, you can use
"*" followed by the array index followed by "$".

The flags characters are a combination of:

    #   Use alternate form.  For "o" conversions, a "0" is prefixed if it is
        not already there.  For "x" conversions, a non-zero result has "0x"
        prepended to it.  For "a", "e", "f", and "g" conversions, the result
        will always contain a decimal point, even if no  digits follow it.
        Additionally for "g" and "G" conversions, trailing zeros are not
        removed from the result.
        
    0   The value should be zero padded.  For "d", "i", "o", "x", "a",
        "e", "f", and "g" conversions, the converted value is padded on the
        left with zeros rather than blanks.  If both "0" and "-" flags appear,
        the "0" flag is ignored.
        
    -   The converted value is to be left-adjusted on the field boundary and
        padded on the right instead.
        
   ' '  (a space) A blank should be left before a positive number or empty
        string produced by a signed conversion.
        
    +   A sign must always be put before a number.  By default a sign is only
        used for negative numbers.  This overrides the " " flag.
        
    '   For "i", "d", "u", "f", and "g", the output is to be grouped according
        to the locale, which may include thousands separators or a
        different decimal character.
        
    I   Use the locale's alternative output digits, if any (for example, Arabic
        digits).
        
    ^   For "x" and "a" conversions, capitalise hexadecimal digits.
        For "g" and "e" conversions, use "E" for the exponent if it's given
        instead of "e".  There are actually two possible instances of this
        flag in "x"; if it occurs before "#", then that means to prepend "0X"
        instead of "0x" with nonzero values, and to capitalise the hex
        digits.  If it occurs after the "#", then it leaves the prefix as
        "0x".
        
The optional field width follows, with a nonzero first digit, and given in
decimal.  If the converted value has fewer characters than the field width, it
will be padded with spaces (or 0, if the "0" flag is given) on the left
(or right, if the "-" flag is given).  Optionally you can provide "*", which
means to consume an argument as an integer to get the field width.  The asterisk
can be followed by an argument index and "$" to take the value from another
argument than what was expected.

The optional precision follows, which is "." followed by a decimal value.  You
can use the same optional forms as with the field width above.  With "d", "i",
"o", "u", and "x" conversions, this specifies the number of digits to be
used (with spaces or zeros used as padding, depending on the "-" and "0" flags).
With "a", "e", and "f" conversions, this is the number of
characters to appear after the radix character.  With "g" conversion,
this is the maximum number of significant digits.  With "s" conversion
on strings, this is the maximum number of characters to be printed.

There is no length modifier as with C printf, as we're always aware of how large
a given type is.

Now then, the conversion specifier.  In general you will not be using anything
other than "%s" and "%r", as these dispatch to the correct type.  But if you 
need to output specific formats, that capability is provided.  Also note that
many C specifiers have not made the transition, with the majority dying due to
the "^" flag.  Here are all the specifiers and their new form:

    %A becomes %^x
    %E becomes %^e
    %F becomes %^f
    %G becomes %^g
    %X becomes %^x
    %a becomes %x
    %i becomes %d
    %u becomes %d
    
Windows has even more dead, due to wchar.  My system doesn't care what you pass
to a string argument; it will handle char[], wchar[], char*, and wchar* equally
well.

Printing out nan, inf, or -inf result in "nan", "inf", or "-inf", unless
otherwise handled by %r.  If the "+" flag is specified, "+inf" is returned
instead.  The "^" flag has no effect on these values.

Printing out complex varies in its interpretation, but the commonality is that
it prints out like "real+imagi".  For example, "1 - 2.4i" formatted with "%d"
comes out as "1-2i".

Real numbers rounded to a precision round up or down as per normal mathematics.
So "3.14159" printed with "%.4f" comes out as "3.1416".
    
These are the specifiers:

    b   Print using binary notation.  Printing reals has correct results.

    d   The numeric argument is converted to decimal notation.  The precision
        gives the minimum number of digits that must appear; if it has fewer
        digits, it's padded.  When 0 is printed with an explicit precision 0,
        the output is empty.
        
        Putting a real value in this argument converts it to decimal, rounding
        in the mathematical manner, so "41.5" is rounded to "42".  This makes it
        equivalent to "%.0f".
        
    e   The argument is rounded and converted to the style "[-]d.ddde+dd", where
        there is one digit before the decimal point and the number of digits
        afterwards depends upon the precision.  If the precision is missing, it
        is taken as 6.  If the precision is zero, no decimal point character
        appears.  This is then followed by "e" or "E" depending upon the "^"
        flag, then "+" or "-" depending upon the exponent, then the exponent.
        The exponent is always at least two digits, zero padded on the left if
        necessary.
        
    f   The real argument is converted to the decimal notation in the style
        "[-]ddd.ddd", where the number of digits after the decimal point is
        equal to the precision specifier.  If the precision is missing, it is
        taken as 6.
        
    g   Use "e" format if the exponent is less than -4 or equal to the
        precision; otherwise "f" is used.  Trailing zeros are removed, as is
        the decimal point if it is not followed by any digits.
        
    o   Print using octal notation.  Printing octal reals has correct results.
    
    s   Print anything.  This dispatches to many of the other formatting
        specifiers, depending on type.  All of the integer types (bit, byte,
        ubyte, short, ushort, int, uint, long, ulong, cent, and ucent) are sent
        to %d.  All of the real types (float, double, extended, and complex) are
        sent to %g.  char[], wchar[], char*, wchar*, char, and wchar are printed
        using this specifier.  That is by merely copying out characters.
        Printing out a wchar value that is out of range (above code 255) in the
        non-wchar forms of fmt prints out nothing.
    
    r   Print using an input representation.  This is like %s, except that:
    
            Values of type bit are printed as "true" or "false".
            
            nan, inf, and -inf are printed as "XXX.nan", "XXX.inf", or
            "-XXX.inf", where "XXX" is "float", "double", "extended", or
            "complex".
            
            Strings are wrapped with double-quotations, and any escapable
            characters in the string are escaped.  If there are no escapable
            characters but backslash and it is shorter or of equivalent
            length, the single-character escape is used instead.
            
            Characters are interpreted as strings, so have the same
            quotations.  They are never written as unquoted escapes.
    
    x   Print using hexadecimal notation.  Printing hexadecimal reals has
        correct results.

*/

module std.format;

import std.ctype;

/* Format the string and return the result */

char [] fmt (char [] format, generic [] args...)
{
    return fmt ("%", format, args...);
}

/* Format text and print out the result */

char [] print (char [] format, generic [] args...)
{
    char [] result = fmt ("%", format, args...);
    
    printf ("%.*s", result);
    return result;
}

/* Format text, print out the result, then a newline */

char [] println (char [] format, generic [] args...)
{
    char [] result = fmt ("%", format, args...);

    printf ("%.*s\n", result);
    return result;
}

/* This variant allows a different introduction specifier than "%".
 * For example, formatting lots of AT&T assembler output would be
 * helped by using "#" here.
 */

static int parseNumber (char [] format, inout int point)
{
    int value = 0;
    
    while (point < format.length && isdigit (format [point]))
        value = value * 10 + (format [point ++] - '0');
    
    return value;
}

char [] fmt (char intro, char [] format, generic [] args...)
{
    int c = 0; /* Current character */
    int p = 0; /* Character after the last format specifier */
    char [] output; /* Output string */
    int end = format.length;
    int current = 0; /* Current argument */
    
    while (1)
    {
        FormatSpecifier spec;
        int readarg = -1; /* Argument index read */
        int widtharg = current;
        
        /* Find the next format specifier and pump out the skipped text */
        p = c;
        while (c < end && format [c] != intro)
            c ++;
        if (c >= end)
            break;
        output ~= format [p .. c];
        
        /* Send out an escaped character */
        if (c == end - 1 || format [c + 1] == intro)
        {
            output ~= intro;
            c += 2;
            continue;
        }
        
        /* Get a new index */
        if (isdigit (format [c + 1]))
        {
            int value = 0;
            int d = c + 1;
            
            while (d < format.length)
            {
                if (format [d ++] == "$")
                {
                    readarg = value - 1;
                    c = d - 1;
                    break;
                }
                
                if (!isdigit (format [d - 1]))
                    break;
                
                value = value * 10 + format [d - 1] - '0';
            }
            
        }
        
        /* Parse the specifier flags */
        while (1)
        {
            c ++;
            if (c >= end)
                return output;
            if (format [c] == "#")
                spec.alternate = true;
            else if (format [c] == "0")
                spec.zero = true;
            else if (format [c] == "-")
                spec.left = true;
            else if (format [c] == " ")
                spec.blank = true;
            else if (format [c] == "+")
                spec.plus = true;
            else if (format [c] == "'")
                spec.locale = true;
            else if (format [c] == "I")
                spec.altdigits = true;
            else if (format [c] == "^")
            {
                if (spec.alternate)
                    spec.precapital = true;
                spec.capital = true;
            }
            else
                break;
        }
        
        /* The field width, if given */
        if (isdigit (format [c]))
            spec.width = parseNumber (format, c);
        else if (format [c] == "*")
        {
            c ++;
            if (readarg == -1)
                readarg = current;
            if (readarg < args.length && readarg >= 0)
                spec.width = args [readarg ++].toInt ();
        }
        
        if (c >= format.length)
            return output;
        
        if (format [c] == ".")
        {
            c ++;
            if (isdigit (format [c]))
                spec.precision = parseNumber (format, c);
            else if (format [c] == "*")
            {
                c ++;
                if (readarg == -1)
                    readarg = current;
                if (readarg < args.length && readarg >= 0)
                    spec.precision = args [readarg ++].toInt ();
            }
        }
        
        spec.code = format [c ++];
        spec.clear ();
        if (readarg == -1)
            readarg = current;
        if (readarg < args.length && readarg >= 0)
            output ~= args [readarg].format (&spec);
        current = readarg + 1;
    }
    
    output ~= format [p .. ];
    return output;
}

static void test (char [] compare, char [] format, generic [] args...)
{
    char [] result;
    
    result = fmt (format, args...);
    if (result != compare)
    {
        printf ("'%.*s' != '%.*s'\n", result, compare);
        assert (result == compare);
    }
}

unittest
{
    test ("Hello, world!", "Hello, world!");
    test ("%%%%x%%", "%%%%%%%%x%%%%");
    test ("45", "%d", 45);
    test ("-35", "%d", -35);
    test ("-52", "%d", -52.345);
    test ("-53", "%d", -52.6);
    test ("+16", "%+d", 16);
    test (" 360", "% d", 360);
    test ("  16", "%4d", 16);
    test (" -32", "%4d", -32);
    test ("0015", "%04d", 15);
    test ("64    ", "%-6d", 64);
    test ("3748", "%s", 3748);
    test ("Hello, world!", "%s, %s!", (char *) "Hello", (char []) "world");
    test ("World, hello!", "%s, %s!", (wchar *) "World", (wchar []) "hello");
    test ("C   .   D", "%-4s.%4s", (char) "C", (wchar) "D");
    test ("true ain't false", "%r ain't %r", true, false);
    test ("'fooodish'", "%r", "fooodish");
    test ("'splar\"f'", "%r", "splar\"f");
    test ('"foo\n"', "%r", "foo\n");
    test ("4.000000", "%f", 4);
    test ("45983", "%.0f", 45983.1);
    test ("46", "%.0f", 45.5);
    test ("16.475000", "%f", (extended) 16.475);
    test ("  0.66667", "%9.5f", 2 / 3.0);
    test ("nan", "%f", float.nan);
    test ("inf", "%f", float.infinity);
    test ("-inf", "%f", -float.infinity);
    test ("4.532000e+01", "%e", 45.32);
    test ("101", "%b", 5);
    test ("43.32", "%g", 43.32);

    //test ("1+2i", "%d", 1+2i);
    
    int [] array = [1, 2, 3, 4, 5];
    test ("[1, 2, 3, 4, 5]", "%s", array);
    
    test ("null", "%s", (char *) null);
    test ("null", "%r", (void *) null);
    test ("00000000", "%p", (void *) null);
    test ("0000beef", "%p", (void *) 48879);
    
    Object obj = new Object ();
    test ("Object", "%s", obj);
    
    test ("2 1 3", "%2$d %1$d %3$d", 1, 2, 3);
    test ("  45", "%*d%d", 3, 4, 5);
}
