/* 

Pickling Module

Pickling, also known as serialisation, marshalling, and flattening, is the
automatic saving and loading of an object.  This module manages this, and
handles pointers and arrays properly.

To start pickling, create a pickler and call dump on it:

    Pickler p = new Pickler ();
    
    p.dump (object);
    
You can save basic type objects as well using a two-argument method:

    int value = 45;

    p.dump (value.typeinfo, &value);
    
The default constructor generates a pickle jar to save to.  You can subclass
the PickleJar to provide your own methods, such as for networking; you only
have to overload the byteSave and byteLoad methods.  The default PickleJar
stores the output data in its array field.  So you can save that using:

    write ("filename.dat", p.jar.array);

To get the objects back, instantiate the Unpickler class with either the
data to read from or a pickle jar to use:

    Unpickler u = new PickleJar (read ("filename.dat"));
    
Then use the load method to read objects:

    object = u.load ();
    
Or a two-argument method to load other things:

    u.load (value.typeinfo, &value);
    
The pickler will only save and load unique objects and arrays once.  If you
want to save the same object changed, you will have to instantiate another
pickler, but this won't modify the previous instance when loading.

Any unusual interactions in arrays will not be retained.  For example, a slice
of another array will be turned into a distinct array on load.  Moreover, it is
assumed that any pointer is pointing to a single object, since we have arrays
and they work very nicely.

The pickler will handle new fields when loading properly.  It will not handle
deleted fields, nor will it handle fields whose type has changed.

*/

module std.pickle;

/* Thrown when we try to read past the end of the pickle jar */

class PickleBoundsError : Error
{
    this (char [] message) { super (message); }
}

/* Thrown when the input stream is bogus */

class PickleBogusError : Error
{
    this (char [] message) { super (message); }
}

/* This is when a symbol can't be found */

class PickleSymbolError : Error
{
    char [] name;
    
    this (SymbolInfo symbol, char [] name)
    {
        this.name = name;
        super ("Module " ~ name ~ " doesn't exist or isn't loaded");
    }
}

/* This is when a saved field doesn't have a matched type to the loading */

class PickleFieldUnmatchedError : Error
{
    TypeInfo save; /* The type saved */
    TypeInfo load; /* The type loaded */
    ClassInfo loadClass; /* The class we're loading for */
    
    this (TypeInfo save, TypeInfo load, ClassInfo name)
    {
        this.save = save;
        this.load = load;
        this.loadClass = name;
        super ("Type " ~ save.toString () ~ ", " ~ save.classinfo.name ~ " and " ~ load.toString () ~ " are incompatible");
    }
}

/* A field with this name wasn't found */

class PickleFieldNotFoundError : Error
{
    char [] name; /* The name of the field */
    TypeInfo type; /* The type of the field */
    ClassInfo info; /* The class we're searching */
    
    this (char [] name, TypeInfo type, ClassInfo info)
    {
        this.name = name;
        this.type = type;
        this.info = info;
        super ("Field " ~ name ~ " wasn't found in class " ~ info.name ~ "!");
    }
}

/* This class handles saving and loading pickle data.  You can overload it to
   save to a stream, for example.  The default implementation simply outputs
   to an array. */

private union Everything
{
    ubyte [16] data;
    float vfloat;
    double vdouble;
    extended vextended;
    int vint;
}

class PickleJar
{
    byte [] array; /* Array to save to/load from */
    int offset; /* For loading, current offset into the array */

    this () { }
    this (byte [] array) { this.array = array; }

    void byteSave (byte [] data) { array ~= data; }
    void byteSave (byte value) { array ~= (byte) value; }

    void ubyteSave (ubyte value) { byteSave (value); }

    /* Save an uncounted set of bytes */
    void ubyteSave (ubyte [] array) { byteSave ((byte []) array); }

    void ulongSave (ulong value)
    {
        do
        {
            if (value > 127)
                ubyteSave ((value & 127) | 128);
            else
                ubyteSave (value & 127);
            value = value >> 7;
        }
        while (value);
    }

    void longSave (long value)
    {
        ulong second;
        
        if (value == long.min)
            throw new Error ("Cannot pickle long.min");
        if (value < 0)
            second = ((-value) << 1) | 1;
        else
            second = value << 1;
        ulongSave (second);
    }

    void shortSave (short value) { longSave (value); }
    void ushortSave (ushort value) { ulongSave (value); }
    void intSave (int value) { longSave (value); }
    void uintSave (uint value) { ulongSave (value); }
    void markerSave (uint index) { uintSave (index); }
    void charSave (char value) { ubyteSave (value); }
    
    void floatSave (float value)
    {
        Everything wrap;
        
        wrap.vfloat = value;
        ubyteSave (wrap.data [0 .. 4]);
    }

    /* This is a counted string */
    void cstringSave (char [] string)
    {
        uintSave (string.length);
        ubyteSave ((ubyte []) string);
    }

    /* -- Loading -- */

    byte byteLoad ()
    {
        if (offset >= array.length)
            throw new PickleBoundsError ("out-of-bounds read on byte");
        return array [offset ++];
    }

    void byteLoad (byte [] data)
    {
        if (offset + data.length > array.length)
            throw new PickleBoundsError ("out-of-bounds read on byte array");
        data [] = array [offset .. offset + data.length];
        offset += data.length;
    }
    
    ulong ulongLoad ()
    {
        ulong value, next;
        int size = 0;
        
        do
        {
            next = ubyteLoad ();
            value = value | ((next & 127) << size);
            size += 7;
        }
        while (next & 128);

        return value;
    }
    
    long longLoad ()
    {
        ulong value;
        
        value = ulongLoad ();
        if (value & 1)
            return -(long) (value >> 1);
        else
            return value >> 1;
    }

    byte ubyteLoad () { return byteLoad (); }
    void ubyteLoad (ubyte [] array) { byteLoad ((byte []) array); }
    short shortLoad () { return longLoad (); }
    ushort ushortLoad () { return ulongLoad (); }
    uint uintLoad () { return ulongLoad (); }
    int intLoad () { return longLoad (); }
    uint markerLoad () { return uintLoad (); }
    char charLoad () { return ubyteLoad (); }

    char [] cstringLoad ()
    {
        char [] string;

        string.length = uintLoad ();
        ubyteLoad ((ubyte []) string);
        return string;
    }
    
    float floatLoad ()
    {
        Everything con;
        
        ubyteLoad (con.data [0 .. 4]);
        return con.vfloat;
    }
}

class Pickler
{
    PickleJar jar;
    TypeInfo [ClassInfo] classes; /* Known class list */
    int [TypeInfo] types; /* Known types list */
    byte [] [] arrays; /* Arrays we know how to save */
    Object [] objects; /* Objects we've saved */

    this ()
    {
        jar = new PickleJar ();
    }

    /* Store the object */
    void dump (Object object)
    {
        dumpClass (object.classinfo); /* Save the type */
        
        /* See if a handle for it already exists */
        for (int c = 0; c < objects.length; c ++)
        {
            if (objects [c] === object)
            {
                jar.uintSave (c + 1);
                return;
            }
        }
            
        /* Not found, output it */
        objects ~= object; /* Register the object */
        jar.uintSave (objects.length); /* Save the handle */
        dumpCoreClass (object.classinfo, object); /* Save the object data */
    }

    /* Store the explicit object */
    void dump (TypeInfo type, void *data)
    {
        dumpType (type);
        dumpCore (type, data);
    }

    /* The innards of the dump */
    void dumpCore (TypeInfo info, void *data)
    {
             if (cast (TypeClass)  info) dump (*(Object *) data);
        else if (cast (TypeArray)  info) dumpCoreArray (cast (TypeArray) info, data);
            
        else if (cast (TypeBit)    info) jar.byteSave   (*(bit *)    data);
        else if (cast (TypeByte)   info) jar.byteSave   (*(byte *)   data);
        else if (cast (TypeUByte)  info) jar.ubyteSave  (*(ubyte *)  data);
        else if (cast (TypeChar)   info) jar.charSave   (*(char *)   data);
        else if (cast (TypeShort)  info) jar.shortSave  (*(short *)  data);
        else if (cast (TypeUShort) info) jar.ushortSave (*(ushort *) data);
        else if (cast (TypeInt)    info) jar.intSave    (*(int *)    data);
        else if (cast (TypeUInt)   info) jar.uintSave   (*(uint *)   data);
        else if (cast (TypeFloat)  info) jar.floatSave  (*(float *)  data);
            
        else throw new Error ("unhandled dump type " ~ info.kind ());
    }

    void dumpCoreArray (TypeArray type, void *data)
    {
        byte [] array = *(byte []*) data;
        int size = type.next.tsize ();
        
        /* If this is null, bail */
        if (array == null)
        {
            jar.uintSave (0);
            return;
        }
        
        /* Find out whether this exists */
        for (int c = 0; c < arrays.length; c ++)
        {
            if ((void *) arrays [c] == (void *) array
             && arrays [c].length == array.length * size)
            {
                jar.uintSave (c + 1);
                return;
            }
        }
            
        /* It doesn't so register a new array */
        jar.uintSave (arrays.length + 1);
        
        byte [] copy = array;
        *(int *) &copy = array.length * size; /* Stash the byte size */
        arrays ~= copy; /* Store the array */
        
        /* First the length */
        jar.uintSave (array.length);

        /* Now each cell */
        for (int c = 0; c < array.length; c ++)
            dumpCore (type.next, (void *) array + c * size);
    }

    void dumpCoreClass (ClassInfo info, void *data)
    {
        /* Dump the base class data if there is one */
        if (info.base)
            dumpCoreClass (info.base, data);

        /* Dump the fields */
        for (int c = 0; c < info.fields.length; c ++)
        {
            FieldInfo field = info.fields [c];

            dumpCore (field.type, data + field.offset);
        }
    }

    /* Store class type, creating if necessary */
    void dumpClass (ClassInfo info)
    {
        if (!info)
            jar.markerSave (0);
        else if (!classes [info])
            dumpType (new TypeClass (info));
        else
            dumpType (classes [info]);
    }

    /* Store a new type */
    void dumpType (TypeInfo type)
    {
        if (type in types)
        {
            jar.markerSave (types [type]);
            return;
        }
        
        types [type] = types.length + 1;
        jar.markerSave (types [type]);

        dumpTypeValue (type);
    }

    /* Dump part of a type */
    void dumpTypeValue (TypeInfo type)
    {
        jar.charSave (type.mangleChar () [0]);
        if (cast (TypeClass) type)
            dumpTypeClass (cast (TypeClass) type);
        else if (cast (TypeArray) type)
            dumpType (type.next); /* Only the "next" is relevant */
        else if (!type.isBasic ())
            throw new Error ("unhandled type kind " ~ type.kind ());
    }

    void dumpSymbolStack (SymbolInfo sym)
    {
        if (sym.parent)
            dumpSymbolStack (sym.parent);
        jar.cstringSave (sym.name);
    }

    void dumpTypeClass (TypeClass type)
    {
        ClassInfo cl = type.sym;
        int depth;

        classes [cl] = type;
        dumpClass (cl.base);
        
        /* Find out how deep the stack is */
        for (SymbolInfo sym = cl.parent; sym; sym = sym.parent)
            depth ++;

        jar.uintSave (depth);

        /* Now dump the stack, as well as this class for ease */
        dumpSymbolStack (cl);

        /* Output the number of fields */
        jar.uintSave (cl.fields.length);
        
        /* Output the fields themselves */
        for (int c = 0; c < cl.fields.length; c ++)
        {
            FieldInfo field = cl.fields [c];
            
            dumpType (field.type);
            jar.cstringSave (field.name);
        }
    }
}

class Unpickler
{
    PickleJar jar; /* Input data */

    TypeInfo [] types; /* List of types we know how to load */
    byte [] [] arrays; /* List of arrays we've loaded */
    Object [] objects; /* List of objects that we've loaded */

    /* Source from data */
    this (byte [] data)
    {
        jar = new PickleJar (data);
    }

    protected TypeInfo nextType; /* Type of the next object waiting or null */
    
    /* Find the next type on the stream */
    TypeInfo next ()
    {
        if (!nextType)
            nextType = loadType ();
        return nextType;
    }
    
    /* Load a class instance object */
    Object load ()
    {
        TypeInfo type = next ();
        ClassInfo info;
        Object obj;
        uint index;
        
        /* Make sure this is a match and then read it */
        if (!cast (TypeClass) type)
            throw new Error ("Next type is not an object, as expected");
        info = (cast (TypeClass) type).sym;
        nextType = null;
        
        index = jar.uintLoad ();
        if (index == 0)
            return null;
        else if (index != objects.length + 1) /* Already present */
        {
            if (index <= 0 || index > objects.length)
                throw new PickleBogusError ("object index is not in range 0 < " ~ itoa (index) ~ " < " ~ itoa (objects.length));
            return objects [index - 1];
        }
        
        obj = info.create ();
        objects ~= obj;
        
        /* Now read in the fields */
        for (int c = 0; c < info.fields.length; c ++)
        {
            FieldInfo field = info.fields [c];
            
            loadCore (field.type, (void *) obj + field.offset);
        }
        
        return obj;
    }
    
    /* Load an object */
    void load (TypeInfo type, void *data)
    {
        TypeInfo info = next ();
        
        if (info != type)
            throw new Error ("Next type is not a " ~ type.toString () ~ " as expected, but a " ~ info.toString ());
        
        nextType = null;
        loadCore (info, data);
    }
    
    /* Load a typed pointer */
    void loadCore (TypeInfo type, void *data)
    {
             if (cast (TypeClass) type) *(Object *) data = load ();
        else if (cast (TypeArray) type) loadArray (cast (TypeArray) type, data);
            
        else if (cast (TypeBit)   type) *(bit *)   data = jar.byteLoad ();
        else if (cast (TypeByte)  type) *(byte *)  data = jar.byteLoad ();
        else if (cast (TypeUByte) type) *(ubyte *) data = jar.ubyteLoad ();
        else if (cast (TypeChar)  type) *(char *)  data = jar.charLoad ();
        else if (cast (TypeShort) type) *(short *) data = jar.shortLoad ();
        else if (cast (TypeUShort)type) *(ushort *)data = jar.ushortLoad (); 
        else if (cast (TypeInt)   type) *(int *)   data = jar.intLoad ();
        else if (cast (TypeUInt)  type) *(uint *)  data = jar.uintLoad ();
        else if (cast (TypeFloat) type) *(float *) data = jar.floatLoad ();
            
        else throw new Error ("Don't know how to handle type " ~ type.kind ());
    }
    
    void loadArray (TypeArray type, void *data)
    {
        int size = type.next.tsize (); /* Size of each cell */
        byte [] array; /* Stored array */
        int length; /* Length of the array */
        uint index; /* Index for loading */
        
        index = jar.uintLoad ();
        if (index == 0) /* Empty */
            *(byte [] *) data = array;
        else if (index != arrays.length + 1) /* Already present */
        {
            if (index <= 0 || index > arrays.length)
                throw new PickleBogusError ("array index is not in range 0 < " ~ itoa (index) ~ " < " ~ itoa (arrays.length));
            array = arrays [index - 1];
        }
        else /* Create a new array */
        {
            /* Load the length and allocate the array */
            length = jar.uintLoad ();
            array.length = length * size + (cast (TypeChar) type.next ? 1 : 0);
        
            /* Load the cells */
            for (int c = 0; c < length; c ++)
                loadCore (type.next, (void *) array + c * size);

            arrays ~= array;
        }
        
        *(int *) &array /= size; /* Stuff the size */
        *(byte []*) data = array; /* Stash the array */
    }

    /* Load a type marker, and possibly a type */
    TypeInfo loadType ()
    {
        char mangle;
        uint index;

        index = jar.uintLoad ();
        if (index == 0)
            return null;
        if (index != types.length + 1)
        {
            if (index <= 0 || index > types.length)
                throw new PickleBogusError ("type index is not in range 0 < " ~ itoa (index) ~ " < " ~ itoa (types.length));
            return types [index - 1];
        }

        /* Get the mangle and figure out where to go from that */
        mangle = jar.charLoad ();

             if (mangle == "A") return loadTypeArray ();
        else if (mangle == "C") return loadTypeClass ();
            
        else if (mangle == "a") return loadTypeBasic (new TypeChar ());
        else if (mangle == "b") return loadTypeBasic (new TypeBit ());
        else if (mangle == "f") return loadTypeBasic (new TypeFloat ());
        else if (mangle == "g") return loadTypeBasic (new TypeByte ());
        else if (mangle == "h") return loadTypeBasic (new TypeUByte ());
        else if (mangle == "i") return loadTypeBasic (new TypeInt ());
        else if (mangle == "k") return loadTypeBasic (new TypeUInt ());
        else if (mangle == "s") return loadTypeBasic (new TypeShort ());
        else if (mangle == "t") return loadTypeBasic (new TypeUShort ());
            
        else
        {
            char [] string;
            
            string = "Unhandled mangle character \"";
            string ~= mangle;
            string ~= "\"";
            throw new PickleBogusError (string);
        }
    }

    /* Read the depth then a set of symbols describing a location */
    
    SymbolInfo findSymbol ()
    {
        SymbolInfo symbol, previous;
        int depth;
        char [] name;

        depth = jar.uintLoad ();
        for (int c = 0; c <= depth; c ++)
        {
            name = jar.cstringLoad ();
            if (!symbol)
                symbol = ModuleInfo.findModule (name);
            else if (cast (ModuleInfo) symbol)
                symbol = (cast (ModuleInfo) symbol).findSymbol (name);
            else
                throw new Error ("find-symbol type " ~ name ~ " " ~ symbol.classinfo.name);

            if (!symbol)
                throw new PickleSymbolError (previous, name);
            previous = symbol;
        }

        return symbol;
    }
    
    TypeInfo loadTypeBasic (TypeInfo type)
    {
        types ~= type; /* Register the type */
        return type; /* Done! */
    }
    
    TypeInfo loadTypeArray ()
    {
        TypeArray type = new TypeArray (); /* The type we're building */
        
        types ~= type; /* Register it */
        type.next = loadType (); /* Load the contained type */
        return type; /* Done! */
    }

    TypeInfo loadTypeClass ()
    {
        ClassInfo info = new ClassInfo (); /* The class we're building */
        ClassInfo compare; /* The actual class we're building for */
        TypeClass type = new TypeClass (info); /* The type we're building */
        SymbolInfo symbol; /* The symbol matching this class */
        TypeInfo base; /* The immediate base class */
        int fieldCount; /* The number of fields to read */

        types ~= type; /* Register it */

        /* Load the base and merge its fields */
        base = loadType ();
        if (base)
        {
            info.base = (cast (TypeClass) base).sym;
            info.fields ~= info.base.fields; /* We take our base class fields */
        }

        /* Find the ClassInfo we're unloading into */
        symbol = findSymbol ();
        if (!cast (ClassInfo) symbol)
            throw new Error ("didn't result in a ClassInfo, but a " ~ symbol.classinfo.name);
        compare = cast (ClassInfo) symbol;
        info.init = compare.init;
        info.name = compare.name;
        info.parent = compare.parent;

        /* Read and match the fields */
        fieldCount = jar.uintLoad ();
        
        for (int c = 0; c < fieldCount; c ++)
        {
            TypeInfo type = loadType ();
            char [] name = jar.cstringLoad ();
            FieldInfo field = new FieldInfo ();
            
            field.name = name;
            field.type = type;
            info.fields ~= field;
            
            /* Try to find a match */
            for (int d = 0; ; d ++)
            {
                if (d >= compare.fields.length)
                    throw new PickleFieldNotFoundError (name, type, compare);
                    
                FieldInfo cmpfield = compare.fields [d];
                
                if (cmpfield.name != name)
                    continue;
                if (cmpfield.type != type)
                    throw new PickleFieldUnmatchedError (type, cmpfield.type, compare);
                field.offset = cmpfield.offset;
                break;
            }
        }
        
        return type;
    }
}
