/* Internal stuff for TypeInfo */

module std.type;

/+
/* This contains all the information about a format conversion. */

struct FormatSpecifier
{
    bit alternate; /* Use alternate form ("#"). */
    bit zero; /* Pad with zeros instead of spaces ("0"). */
    bit left; /* Left-adjust and pad on the right instead ("-"). */
    bit blank; /* Put a blank before a positive number or empty string (" "). */
    bit plus; /* Always place a sign before a number ("+"). */
    bit locale; /* Use locale printing and thousands separators ("'"). */
    bit altdigits; /* Use the locale's alternate output digits, if any ("I"). */
    bit capital; /* Capitalise hex digits ("^"). */
    bit precapital; /* Capitalise the "0x" prefix ("^" before "#"). */
    
    int width = -1; /* Field width, negative if unspecified. */
    int precision = -1; /* Field precision, negative if unspecified. */

    char code; /* Actual specifier */
}
+/

static char [] digitsLow = "0123456789abcdef";
static char [] digitsHigh = "0123456789ABCDEF";

import std.string;

static char [] formatNumber (FormatSpecifier *spec, char [] text, bit signed)
{
    bit hasSigned = signed || spec.blank || spec.plus;
    bit padded = false;
    
    if (spec.zero && text.length < spec.width)
    {
        padded = true;
        text = std.string.repeat ("0", spec.width - text.length) ~ text;
    }
    
    if (signed)
        text = "-" ~ text;
    else if (spec.blank)
        text = " " ~ text;
    else if (spec.plus)
        text = "+" ~ text;
    
    if (!padded && text.length < spec.width)
    {
        padded = true;
        if (spec.left)
            text ~= std.string.repeat (" ", spec.width - text.length);
        else
            text = std.string.repeat (" ", spec.width - text.length) ~ text;
    }
    
    return text;
}

static char [] formatLong (FormatSpecifier *spec, ulong value, bit signed)
{
    char *digits = (spec.capital ? digitsHigh : digitsLow);
    int radix = spec.radix (), digit;
    char [] text;
    
    if (spec.isReal ())
        return formatReal (spec, signed ? -(extended) value : value);
    
    do
    {
        digit = value % radix;
        text = digits [digit .. digit + 1] ~ text;
        value /= radix;
    }
    while (value);
        
    return formatNumber (spec, text, signed);
}

static char [] formatBignum (FormatSpecifier *spec, bignum value)
{
    char *digits = (spec.capital ? digitsHigh : digitsLow);
    int radix = spec.radix (), digit;
    char [] text;
    bit signed;
    
    if (value < 0)
        signed = true, value = -value;

    do
    {
        digit = value % radix;
        text = digits [digit .. digit + 1] ~ text;
        value /= radix;
    }
    while (value);
    
    return formatNumber (spec, text, signed);
}

import std.math;

static char [] formatReal (FormatSpecifier *spec, extended value)
{
    int exp;
    
    return formatReal (spec, value, exp, spec.radix ());
}

static char [] formatReal (FormatSpecifier *spec, extended value, out int exporig, int radix)
{
    if (value == extended.infinity || value == -extended.infinity)
        return formatNumber (spec, "inf", value < 0);
    else if (value.isnan)
        return formatNumber (spec, "nan", false);
    
    extended orig = value;
    char *digits = (spec.capital ? digitsHigh : digitsLow);
    int precision = (spec.precision < 0 ? 6 : spec.precision);
    bit signed = value < 0;
    int digit, c;
    char [] text;
    
    value = (value < 0 ? -value : value);
    
    if (spec.code == "f" || spec.code == "d")
    {
        extended dig;
        
        if (spec.code == "d")
            precision = 0;
        
        if (precision == 0)
            value = floor (value + 0.5);
        
        do
        {
            digit = value % radix;
            text = digits [digit .. digit + 1] ~ text;
            value /= radix;
        }
        while (value >= 1);

        if (precision)
            text ~= ".";
        value = orig - floor (orig);
        for (c = 0; c < precision; c ++)
        {
            dig = value * pow (radix, c + 1);
            digit = (c < precision - 1 ? dig % radix : (dig + 0.5) % radix);
            text ~= digits [digit .. digit + 1];
        }
        
        return formatNumber (spec, text, signed);
    }
    
    if (spec.code == "e")
    {
        extended dig;
        int exp; /* Exponent */
        int count;
        char [] subtext;
        
        if (value >= 1)
        {
            for (exp = 0; pow (10, exp + 1) < value; exp ++)
            {
            }
        }
        else if (value < 1)
        {
            for (exp = 0; value > pow (10, exp - 1); exp --)
                { }
        }

        exporig = exp;
        orig = value = value * pow (10, -exp);
        
        if (precision == 0)
            value = floor (value + 0.5);
        
        do
        {
            digit = value % radix;
            text = digits [digit .. digit + 1] ~ text;
            value /= radix;
        }
        while (value >= 1);
            
        if (precision)
            text ~= ".";
        value = orig;// - floor (orig);
        for (c = 0; c < precision; c ++)
        {
            dig = value * pow (radix, c + 1);
            digit = (c < precision - 1 ? dig % radix : (dig + 0.5) % radix);
            text ~= digits [digit .. digit + 1];
        }
        
        text ~= spec.capital ? "E" : "e";
        
        if (exp < 0)
            text ~= "-", exp = -exp;
        else
            text ~= "+";
        
        do
        {
            digit = exp % 10;
            exp /= 10;
            subtext = digits [digit .. digit + 1] ~ subtext;
        }
        while (exp || count ++ < 1);
            
        text ~= subtext;
        
        return formatNumber (spec, text, signed);
    }
    
    char code = spec.code;
    int exp;
    
    spec.code = "e";
    text = formatReal (spec, value, exp, radix);
    if (exp < 0 ? exp > -4 : exp < 4)
    {
        spec.code = "f";
        text = formatReal (spec, value, exp, radix);
    }
    
    while (text.length && text [text.length - 1] == "0")
        text = text [.. text.length - 1];
    spec.code = code;
    return text;
}

static char [] formatChars (FormatSpecifier *spec, char [] text)
{
    if (spec.code == "r")
    {
        int backslashes, escapes;
        int c;
        
        /* Figure out what we're going to use */
        for (c = 0; c < text.length; c ++)
            if (text [c] == "\\")
                backslashes ++;
            else if (text [c] < 32 || text [c] == "'")
            {
                escapes ++;
                break;
            }
            
        /* There was an escape, so we must use double-quoting */
        if (escapes)
        {
            char [] output;
            
            output = "\"";
            for (c = 0; c < text.length; c ++)
            {
                if (text [c] == "\\" || text [c] == "\"")
                    output ~= "\\" ~ text [c .. c + 1];
                else if (text [c] == "\a") output ~= "\\a";
                else if (text [c] == "\b") output ~= "\\b";
                else if (text [c] == "\f") output ~= "\\f";
                else if (text [c] == "\n") output ~= "\\n";
                else if (text [c] == "\r") output ~= "\\r";
                else if (text [c] == "\t") output ~= "\\t";
                else if (text [c] == "\v") output ~= "\\v";
                else if (text [c] < 32) output ~= "\\x" ~ fmt ("%02X", text [c]);
                else output ~= text [c];
            }
            
            text = output ~ "\"";
        }
        /* Single-quoting is our guide */
        else
            text = "'" ~ text ~ "'";
    }
    
    if (text.length < spec.width)
    {
        if (spec.left)
            text ~= std.string.repeat (" ", spec.width - text.length);
        else
            text = std.string.repeat (" ", spec.width - text.length) ~ text;
    }
    
    return text;
}

/* -- void --------------------------------------------------------------- */

class TypeVoid : TypeInfo
{
override:
    int equals (void *a, void *b) { return 1; }
    uint getHash (void *p) { return 0; }
    int compare (void *a, void *b) { return 0; }
    
    int tsize () { return 0; }
    char [] kind () { return "void"; }
    char [] mangleChar () { return "v"; } 
    bit isBasic () { return true; }
}

/* -- bit ---------------------------------------------------------------- */

static bit bitInit;

class TypeBit : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(bit *) a == *(bit *) b; }
    uint getHash (void *p) { return *(bit *) p; }
    int compare (void *a, void *b) { return (int) *(bit *) a - (int) *(bit *) b; }
    void swap (void *a, void *b) { bit mid = *(bit *) a; *(bit *) a = *(bit *) b; *(bit *) b = mid; }
    int tsize () { return bit.size; }
    char [] kind () { return "bit"; }
    bit toBit (void *a) { return *(bit *) a; }
    int toInt (void *a) { return *(bit *) a; }
    char [] mangleChar () { return "b"; }
    bit isBasic () { return true; }
    byte [] tinit () { return ((byte *) &bitInit) [0 .. bit.size]; }
        
    char [] format (FormatSpecifier *spec, void *data)
    {
        if (spec.code == "r")
        {
            char [] result;
            
            spec.code = "s";
            result = formatChars (spec, *(bit *) data ? "true" : "false");
            spec.code = "r";
            return result;
        }
            printf ("%d\n", *(bit *) data);
        return formatLong (spec, *(bit *) data, false);
    }
}

/* -- byte --------------------------------------------------------------- */

static byte byteInit;

class TypeByte : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(byte *) a == *(byte *) b; }
    uint getHash (void *p) { return *(byte *) p; }
    int compare (void *a, void *b) { return (int) *(byte *) a - (int) *(byte *) b; }
    void swap (void *a, void *b) { byte mid = *(byte *) a; *(byte *) a = *(byte *) b; *(byte *) b = mid; }
    int tsize () { return byte.size; }
    char [] kind () { return "byte"; }
    bit toBit (void *a) { return *(byte *) a; }
    int toInt (void *a) { return *(byte *) a; }
    char [] mangleChar () { return "g"; }
    bit isBasic () { return true; }
    byte [] tinit () { return ((byte *) &byteInit) [0 .. byte.size]; }    
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        int value = *(byte *) data;
        
        return formatLong (spec, value < 0 ? -value : value, value < 0);
    }
}

/* -- ubyte -------------------------------------------------------------- */

static ubyte ubyteInit;

class TypeUByte : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(ubyte *) a == *(ubyte *) b; }
    uint getHash (void *p) { return *(ubyte *) p; }
    int compare (void *a, void *b) { return (int) *(ubyte *) a - (int) *(ubyte *) b; }
    void swap (void *a, void *b) { ubyte mid = *(ubyte *) a; *(ubyte *) a = *(ubyte *) b; *(ubyte *) b = mid; }
    int tsize () { return ubyte.size; }
    char [] kind () { return "ubyte"; }
    bit toBit (void *a) { return *(ubyte *) a; }
    int toInt (void *a) { return *(ubyte *) a; }
    char [] mangleChar () { return "h"; }
    bit isBasic () { return true; }
    byte [] tinit () { return ((byte *) &ubyteInit) [0 .. ubyte.size]; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        int value = *(ubyte *) data;
        
        return formatLong (spec, value, false);
    }
}

/* -- char --------------------------------------------------------------- */

static char charInit;

class TypeChar : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(char *) a == *(char *) b; }
    uint getHash (void *p) { return *(char *) p; }
    int compare (void *a, void *b) { return *(char *) a - *(char *) b; }
    void swap (void *p1, void *p2) { char t; t = *(char *) p1; *(char *) p1 = *(char *) p2; *(char *) p2 = t; }
    int tsize () { return char.size; }
    char [] kind () { return "char"; }
    bit toBit (void *a) { return *(char *) a ? true : false; }
    int toInt (void *a) { return *(char *) a; }
    char [] mangleChar () { return "a"; }
    byte [] tinit () { return ((byte *) &charInit) [0 .. char.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        return formatChars (spec, ((char *) data) [0 .. 1]);
    }
}

/* -- wchar -------------------------------------------------------------- */

static wchar wcharInit;

class TypeWChar : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(wchar *) a == *(wchar *) b; }
    uint getHash (void *p) { return *(wchar *) p; }
    int compare (void *a, void *b) { return *(wchar *) a - *(wchar *) b; }
    void swap (void *p1, void *p2) { wchar t; t = *(wchar *) p1; *(wchar *) p1 = *(wchar *) p2; *(wchar *) p2 = t; }
    int tsize () { return wchar.size; }
    char [] kind () { return "wchar"; }
    bit toBit (void *a) { return *(wchar *) a ? true : false; }
    int toInt (void *a) { return *(wchar *) a; }
    char [] mangleChar () { return "u"; }
    byte [] tinit () { return ((byte *) &wcharInit) [0 .. wchar.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        return formatChars (spec, (char []) ((char *) data) [0 .. 1]);
    }
}

/* -- short -------------------------------------------------------------- */

static short shortInit;

class TypeShort : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(short *) a == *(short *) b; }
    uint getHash (void *p) { return *(short *) p; }
    int compare (void *a, void *b) { return *(short *) a - *(short *) b; }
    void swap (void *p1, void *p2) { int mid; mid = *(short *) p1; *(short *) p1 = *(short *) p2; *(short *) p2 = mid; }
    int tsize () { return short.size; }
    char [] kind () { return "short"; }
    bit toBit (void *a) { return *(short *) a ? true : false; }
    int toInt (void *a) { return *(short *) a; }
    char [] mangleChar () { return "s"; }
    byte [] tinit () { return ((byte *) &shortInit) [0 .. short.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        int value = *(short *) data;
        
        return formatLong (spec, value < 0 ? -value : value, value < 0);
    }
}

/* -- ushort ------------------------------------------------------------- */

static ushort ushortInit;

class TypeUShort : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(ushort *) a == *(ushort *) b; }
    uint getHash (void *p) { return *(ushort *) p; }
    int compare (void *a, void *b) { return *(ushort *) a - *(ushort *) b; }
    void swap (void *p1, void *p2) { int mid; mid = *(ushort *) p1; *(ushort *) p1 = *(ushort *) p2; *(ushort *) p2 = mid; }
    int tsize () { return ushort.size; }
    char [] kind () { return "ushort"; }
    bit toBit (void *a) { return *(ushort *) a ? true : false; }
    int toInt (void *a) { return *(ushort *) a; }
    char [] mangleChar () { return "t"; }
    byte [] tinit () { return ((byte *) &ushortInit) [0 .. ushort.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        return formatLong (spec, *(ushort *) data, false);
    }
}

/* -- int ---------------------------------------------------------------- */

static int intInit;

class TypeInt : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(int *) a == *(int *) b; }
    uint getHash (void *p) { return *(int *) p; }
    int compare (void *a, void *b) { return *(int *) a - *(int *) b; }
    void swap (void *p1, void *p2) { int mid; mid = *(int *) p1; *(int *) p1 = *(int *) p2; *(int *) p2 = mid; }
    int tsize () { return int.size; }
    char [] kind () { return "int"; }
    bit toBit (void *a) { return *(int *) a ? true : false; }
    int toInt (void *a) { return *(int *) a; }
    char [] mangleChar () { return "i"; }
    byte [] tinit () { return ((byte *) &intInit) [0 .. int.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        int value = *(int *) data;
        
        return formatLong (spec, value < 0 ? -value : value, value < 0);
    }
}

/* -- uint --------------------------------------------------------------- */

static uint uintInit;

class TypeUInt : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(uint *) a == *(uint *) b; }
    uint getHash (void *p) { return *(uint *) p; }
    int compare (void *a, void *b) { return *(uint *) a - *(uint *) b; }
    void swap (void *a, void *b) { uint c; c = *(uint *) a; *(uint *) a = *(uint *) b; *(uint *) b = c; }
    int tsize () { return 4; }
    char [] kind () { return "uint"; }
    bit toBit (void *a) { return *(uint *) a ? true : false; }
    int toInt (void *a) { return *(uint *) a; }
    char [] mangleChar () { return 'k'; }
    byte [] tinit () { return ((byte *) &uintInit) [0 .. 4]; }
    bit isBasic () { return true; }
    char [] format (FormatSpecifier *spec, void *data) { return formatLong (spec, *(uint *) data, false); }
}

/* -- long --------------------------------------------------------------- */

static long longInit;

class TypeLong : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(long *) a == *(long *) b; }
    uint getHash (void *p) { return *(long *) p; }
    int compare (void *a, void *b) { return *(long *) a < *(long *) b ? -1 : *(long *) a > *(long *) b ? 1 : 0; }
    void swap (void *a, void *b) { long c; c = *(long *) a; *(long *) a = *(long *) b; *(long *) b = c; }
    int tsize () { return long.size; }
    char [] kind () { return "long"; }
    bit toBit (void *a) { return *(long *) a; }
    int toInt (void *a) { return *(long *) a; }
    char [] mangleChar () { return 'l'; }
    byte [] tinit () { return ((byte *) &longInit) [0 .. 8]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        long value = *(long *) data;
        
        return formatLong (spec, value < 0 ? -value : value, value < 0);
    }
}

/* -- ulong -------------------------------------------------------------- */

static ulong ulongInit;

class TypeULong : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(ulong *) a == *(ulong *) b; }
    uint getHash (void *p) { return *(ulong *) p; }
    int compare (void *a, void *b) { return *(ulong *) a < *(ulong *) b ? -1 : *(ulong *) a > *(ulong *) b ? 1 : 0; }
    void swap (void *a, void *b) { long c; c = *(ulong *) a; *(ulong *) a = *(ulong *) b; *(ulong *) b = c; }
    int tsize () { return ulong.size; }
    char [] kind () { return "ulong"; }
    bit toBit (void *a) { return *(ulong *) a; }
    int toInt (void *a) { return *(ulong *) a; }
    char [] mangleChar () { return 'm'; }
    byte [] tinit () { return ((byte *) &ulongInit) [0 .. ulong.size]; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        ulong value = *(ulong *) data;
        
        return formatLong (spec, value, false);
    }
}

/* -- bignum ------------------------------------------------------------- */

static bignum bignumInit;

class TypeBignum : TypeInfo
{
override:
    char [] format (FormatSpecifier *spec, void *data) { return formatBignum (spec, *(bignum *) data); }
}

/* -- float -------------------------------------------------------------- */

static float floatInit;

class TypeFloat : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(float *) a == *(float *) b; }
    uint getHash (void *p) { return *(float *) p; }
    int tsize () { return 4; }
    char [] kind () { return "float"; }
    bit toBit (void *a) { return *(float *) a ? true : false; }
    int toInt (void *a) { return *(float *) a; }
    char [] mangleChar () { return "f"; }
    void swap (void *pa, void *pb) { float mid = *(float *) pa; *(float *) pa = *(float *) pb; *(float *) pb = mid; }
    byte [] tinit () { return ((byte *) &floatInit) [0 .. 4]; }
    int compare (void *a, void *b) { return *(float *) a < *(float *) b ? -1 : *(float *) a > *(float *) b ? 1 : 0; }
    bit isReal () { return true; }
    bit isBasic () { return true; }
    char [] format (FormatSpecifier *spec, void *data) { return formatReal (spec, *(float *) data); }
}

/* -- double ------------------------------------------------------------- */

static double doubleInit;

class TypeDouble : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(double *) a == *(double *) b; }
    uint getHash (void *p) { return *(double *) p; }
    int compare (void *a, void *b) { return *(double *) a < *(double *) b ? -1 : *(double *) a > *(double *) b ? 1 : 0; }
    int tsize () { return double.size; }
    char [] kind () { return "double"; }
    void swap (void *pa, void *pb) { double mid; mid = *(double*) pa; *(double *) pa = *(double *) pb; *(double *) pb = mid; }
    bit isReal () { return true; }
    bit toBit (void *a) { return *(double *) a ? true : false; }
    int toInt (void *a) { return *(double *) a; }
    byte [] tinit () { return ((byte *) &doubleInit) [0 .. double.size]; }
    char [] mangleChar () { return "d"; }
    bit isBasic () { return true; }
    char [] format (FormatSpecifier *spec, void *data) { return formatReal (spec, *(double *) data); }
}

/* -- complex ------------------------------------------------------------ */

static complex complexInit;

class TypeComplex : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(complex *) a == *(complex *) b; }
    uint getHash (void *p) { complex v = *(complex *) p; return v.re + v.im; }
    //int compare (complex *a, complex *b) { return *(complex *) a < *(complex *) b ? -1 : *(complex *) a > *(complex *) b ? 1 : 0; }
    int tsize () { return complex.size; }
    char [] kind () { return "complex"; }
    void swap (void *pa, void *pb) { complex mid; mid = *(complex *) pa; *(complex *) pa = *(complex *) pb; *(complex *) pb = mid; }
    bit isReal () { return true; }
    //bit toBit (void *a) { return *(complex *) a ? true : false; }
    //int toInt (void *a) { return *(complex *) a; }
    byte [] tinit () { return ((byte *) &complexInit) [0 .. complex.size]; }
    char [] mangleChar () { return "c"; }
    bit isBasic () { return true; }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        complex value = *(complex *) data;
        FormatSpecifier imag = *spec;
        char [] text, add;
        
        
        text = formatReal (spec, value.re);
        imag.plus = true;
        imag.blank = false;
        add = formatReal (&imag, value.im);
        return text ~ add ~ "i";
    }
}

/* -- extended ----------------------------------------------------------- */

static extended extendedInit;

class TypeExtended : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(extended *) a == *(extended *) b; }
    uint getHash (void *p) { return *(extended *) p; }
    int compare (void *a, void *b) { return *(extended *) a < *(extended *) b ? -1 : *(extended *) a > *(extended *) b ? 1 : 0; }
    int tsize () { return extended.size; }
    char [] kind () { return "extended"; }
    void swap (void *pa, void *pb) { extended mid; mid = *(extended *) pa; *(extended *) pa = *(extended *) pb; *(extended *) pb = mid; }
    bit isReal () { return true; }
    bit toBit (void *a) { return *(extended *) a ? true : false; }
    int toInt (void *a) { return *(extended *) a; }
    byte [] tinit () { return ((byte *) &extendedInit) [0 .. extended.size]; }
    char [] mangleChar () { return "e"; }
    bit isBasic () { return true; }
    char [] format (FormatSpecifier *spec, void *data) { return formatReal (spec, *(extended *) data); }
}

/* -- pointer ------------------------------------------------------------ */

static void *pointerInit;

class TypePointer : TypeInfo
{
override:
    int equals (void *a, void *b) { return *(void **) a == *(void **) b; }
    uint getHash (void *p) { return (uint) *(void **) p; }
    int compare (void *a, void *b) { return *(void **) a - *(void **) b; }
    int tsize () { return 4; }
    char [] kind () { return "pointer"; }
    void swap (void *a, void *b) { void *mid = *(void **) a; *(void **) a = *(void **) b; *(void **) b = mid; }
    byte [] tinit () { return ((byte *) &pointerInit) [0 .. 4]; }
    char [] mangleChar () { return "P"; }
    char [] mangle () { return next ? mangleChar () ~ next.mangle () : mangleChar (); }

    char [] toString (char [] name)
    {
        char [] text;
        
        if (next)
        {
            if (cast (TypeFunction) (next))
                return (cast (TypeFunction) (next)).toString (name, true);
            else if (cast (TypeArray) (next) || cast (TypePointer) (next))
                text = next.toString () ~ "*";
            else
                text = next.toString () ~ " *";

            if (name)
                text ~= name;
            return text;
        }
        else
            return kind ();
    }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        void *point = *(void **) data;
        
        if (spec.code == "p")
        {
            FormatSpecifier copy = *spec;
            
            copy.width = 8;
            copy.zero = true;
            copy.code = "x";
            return formatLong (&copy, (uint) point, false);
        }
        
        if (point == null)
        {
            char code = spec.code;
            char [] text;
                
            spec.code = "s";
            text = formatChars (spec, "null");
            spec.code = code;
            return text;
        }
        
        if (cast (TypeChar) next)
        {
            char *pointer = *(char **) data;
            
            char [] string = pointer [0 .. strlen (pointer)];
            
            return formatChars (spec, string);
        }
        
        if (cast (TypeWChar) next)
        {
            wchar *pointer = *(wchar **) data;
            int length;
            
            for (length = 0; pointer [length]; length ++)
                { }
                
            return formatChars (spec, (char []) pointer [0 .. length]);
        }
        
        throw new Error ("can't handle this type of pointer");
    }
}

/* -- array -------------------------------------------------------------- */

import crc32;

class TypeArray : TypeInfo
{
    /* The type in "next" is what is contained by the array. */
    
    this () { }
    this (TypeInfo next) { this.next = next; }
    
override:
    int equals (void *a, void *b) { throw new Error ("Foo equals"); return 0; }
    
    uint getHash (void *p)
    {
        byte [] array = *(byte []*) p;
        uint hash = init_crc32 (), then;
        int s = next.tsize ();
        int c;

        if (cast (TypeChar) next)
        {
            char [] foo = (char []) array;
            
            for (c = 0; c < array.length; c ++)
                hash = update_crc32 (foo [c], hash);
        }
        else
            for (c = 0; c < array.length; c ++)
            {
                then = next.getHash ((byte *) array + c * s);
                hash = update_crc32 ((ubyte) then, hash);
                hash = update_crc32 ((ubyte) (then >> 8), hash);
                hash = update_crc32 ((ubyte) (then >> 16), hash);
                hash = update_crc32 ((ubyte) (then >> 24), hash);
            }
       
        return hash & int.max;
    }
    
    int compare (void *pa, void *pb)
    {
        byte [] a = *(byte []*) pa;
        byte [] b = *(byte []*) pb;
        int cl = a.length < b.length ? a.length : b.length;
        int s = next.tsize ();
        int c, value;

        for (c = 0; c < cl; c ++)
        {
            value = next.compare ((byte *) a + c * s, (byte *) b + c * s);
            if (value)
                return value;
        }

        if (a.length < b.length)
            return -1;
        if (a.length > b.length)
            return 1;
        return 0;
    }

    int tsize () { return 8; }
    char [] kind () { return "array"; }
    char [] mangleChar () { return "A"; }
    char [] mangle () { return next ? mangleChar () ~ next.mangle () : mangleChar (); }

    void swap (void *pa, void *pb)
    {
        int mida, midb;

        mida = ((int *) pa) [0];
        midb = ((int *) pa) [1];

        ((int *) pa) [0] = ((int *) pb) [0];
        ((int *) pa) [1] = ((int *) pb) [1];

        ((int *) pb) [0] = mida;
        ((int *) pb) [1] = midb;
    }

    byte [] tinit () { return new byte [8]; }

    char [] toString (char [] name)
    {
        if (next)
        {
            if (cast (TypePointer) (next) || cast (TypeArray) (next))
                return next.toString (name) ~ "[]";
            return next.toString (name) ~ " []";
        }
        else
            return kind ();
    }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        if (cast (TypeChar) next)
            return formatChars (spec, *(char []*) data);
        if (cast (TypeWChar) next)
            return formatChars (spec, (char []) *(wchar []*) data);
        
        byte [] list = *(byte []*) data;
        int size = next.tsize ();
        char [] text;
        int off, end = size * list.length;
        
        text = "[";
        for (off = 0; off < end; off += size)
        {
            if (off)
                text ~= ", ";
            text ~= next.format (spec, (byte *) list + off);
        }
        return text ~ "]";
    }
}

/* -- sarray --------------------------------------------------------------- */

static import std.string;

class TypeSArray : TypeInfo
{
    int dim; /* Size of the array */
    
override:
    int equals (void *a, void *b) { throw new Error ("Foo equals"); return 0; }
    uint getHash (void *p) { throw new Error ("Foo getHash"); return 0; }
    int compare (void *a, void *b) { throw new Error ("Foo compare"); return 0; }
    int tsize () { return 8; }
    char [] kind () { return "sarray"; }
    
    void swap (void *pa, void *pb) { throw new Error ("foo swap"); }
    byte [] tinit () { throw new Error ("Fooooo"); return new byte [1]; }
    char [] mangleChar () { return "G"; }
    
    char [] toString (char [] name)
    {
        if (next)
            return next.toString (name) ~ " [" ~ itoa (dim) ~ "]";
        else
            return kind ();
    }
}

/* -- aarray ------------------------------------------------------------- */

class TypeAArray : TypeInfo
{
override:
    int equals (void *a, void *b) { throw new Error ("Foo equals"); return 0; }
    uint getHash (void *p) { throw new Error ("Foo getHash"); return 0; }
    int compare (void *a, void *b) { throw new Error ("Foo compare"); return 0; }
    int tsize () { return 8; }
    char [] kind () { return "aarray"; }
    
    void swap (void *pa, void *pb) { throw new Error ("foo swap"); }
    byte [] tinit () { throw new Error ("Fooooo"); return new byte [1]; }
    char [] mangleChar () { return "H"; }
    
    char [] toString (char [] name)
    {
        return kind ();
    }
}

/* -- enum --------------------------------------------------------------- */

class TypeEnum : TypeInfo
{
    TypeInfo memtype; /* Type of the members */
    long maxval; /* Maximum value */
    long minval; /* Minimum value */
    long defval; /* Default value */

    struct Value
    {
        long value;
        char [] name;
    }

    Value [] values; /* List of values */

    /* Return whether this value exists in the enumeration */
    bit hasValue (long val)
    {
        for (int c = 0; c < values.length; c ++)
            if (values [c].value == val)
                return true;
        return false;
    }

    /* Get the name of a value, return "" if it is not in the enumeration */
    char [] valueName (long val)
    {
        for (int c = 0; c < values.length; c ++)
            if (values [c].value == val)
                return values [c].name;
        return "";
    }
    
override:
    int equals (void *a, void *b) { return memtype.equals (a, b); }
    uint getHash (void *p) { return memtype.getHash (p); }
    int compare (void *a, void *b) { return memtype.compare (a, b); }
    int tsize () { return memtype.tsize (); }
    char [] kind () { return "enum"; }

    void swap (void *pa, void *pb) { memtype.swap (pa, pb); }
    byte [] tinit () { return memtype.tinit (); }
    char [] mangleChar () { return "E"; }

    char [] toString (char [] name)
    {
        char [] text;

        text ~= "enum";
        if (memtype != int.typeinfo)
            text ~= " : " ~ memtype.kind ();

        return text;
    }
}

/* -- function ----------------------------------------------------------- */

private void *functionInit;

class TypeFunction : TypeInfo
{
    struct Argument
    {
        TypeInfo type; /* Type of the argument */
        char [] ident; /* Assigned identifier, empty if none */
        char setIn; /* Set if "in" was given */
        char setOut; /* Set if "out" was given */
        char setInOut; /* Set if "inout" was given */
    }

    enum Linkage : int
    {
        Default,
        D,
        C,
        CPP,
        Windows,
        Pascal,
    }
    
    Argument [] args; /* Arguments to the function */
    char varargs; /* Set if this is a varargs function */

    Linkage linkage; /* Linkage type */

    char [] linkageText () { return linkageText (linkage); }

    char [] linkageText (Linkage value)
    {
        switch (value)
        {
            case Linkage.Default: return "extern (Default)";
            case Linkage.D: return "extern (D)";
            case Linkage.C: return "extern (C)";
            case Linkage.CPP: return "extern (CPP)";
            case Linkage.Windows: return "extern (Windows)";
            case Linkage.Pascal: return "extern (Pascal)";
        }
    }

    char [] toString (char [] name, bit isPointer)
    {
        char [] text;

        if (linkage != Linkage.D)
            text ~= linkageText () ~ " ";

        if (next)
        {
            text ~= next.toString ();
            if (!cast (TypePointer) (next))
                text ~= " ";
        }

        if (isPointer)
            text ~= "(*" ~ name ~ ")";
        else
            text ~= name;
            
        text ~= " (";

        for (int a = 0; a < args.length; a ++)
        {
            Argument *arg = &args [a];

            if (a) text ~= ", ";
            text ~= arg.type.toString (arg.ident);
        }

        if (varargs)
        {
            if (args.length) text ~= ", ";
            text ~= "...";
        }
        
        text ~= ")";
        
        return text;
    }
    
override:
    int equals (void *a, void *b) { return *(void **) a === *(void **) b; }
    uint getHash (void *a) { return *(int *) a; }
    int compare (void *a, void *b) { return *(int *) a - *(int *) b; }
    int tsize () { return 4; }
    char [] kind () { return "function"; }
    byte [] tinit () { return ((byte *) &functionInit) [0 .. 4]; }
    char [] mangleChar () { return "F"; }

    char [] mangle ()
    {
        char [] text;

        text = mangleChar ();
        for (int a = 0; a < args.length; a ++)
            text ~= args [a].type.mangle ();
        if (varargs) text ~= "Y"; else text ~= "Z";
        if (next) text ~= next.mangle ();
        return text;
    }

    char [] toString (char [] name) { return toString (name, 0); }
}

/* -- class -------------------------------------------------------------- */

private Object classInit;
private Object [] classRecurse; /* Recursion control for formatting */

class TypeClass : TypeInfo
{
    ClassInfo sym; /* Class this type is for */

    /* A note.  obj.typeinfo.sym is not the same as obj.classinfo.
     * obj.typeinfo.sym is the class the container is defined for;
     * obj.classinfo is the class of the contained object.
     */

    this () { }
    this (ClassInfo sym) { this.sym = sym; }
    
override:
    int equals (void *a, void *b) { return (*(Object *) a).eq (*(Object *) b); }
    uint getHash (void *p) { return (*(Object *) p).toHash (); }
    int compare (void *a, void *b) { return (*(Object *) a).cmp (*(Object *) b); }
    int tsize () { return 4; }
    char [] kind () { return sym.name; }
    void swap (void *a, void *b) { void *m = *(void **) a; *(void **) a = *(void **) b; *(void **) b = m; }
    byte [] tinit () { return ((byte *) &classInit) [0 .. 4]; }
    char [] mangleChar () { return "C"; }

    char [] toString ()
    {
        return sym.name;
    }
    
    int eq (Object foo)
    {
        TypeClass cmp = cast (TypeClass) foo;
        
        if (!cmp)
            return false;
        
        if (sym != cmp.sym)
            return false;
        
        return true;
    }
    
    char [] format (FormatSpecifier *spec, void *data)
    {
        Object obj = *(Object *) data;
        char [] text;
        
        if (obj == null)
            return formatChars (spec, "null");
        
        for (int c = 0; c < classRecurse.length; c ++)
        {
            if (classRecurse [c] == obj)
                return formatChars (spec, "...");
        }
        
        classRecurse ~= obj;
        try
            text = obj.toFormat (spec);
        finally
            classRecurse.length = classRecurse.length - 1;
        
        return text;
    }
}

/* -- struct ------------------------------------------------------------- */

class TypeStruct : TypeInfo
{
    StructInfo sym;
    
override:
    char [] kind () { return "struct"; }
    
    byte [] tinit () { return sym.init; }
    int tsize () { return sym.init.length; }
    
    int equals (void *p1, void *p2) { return ((ubyte *) p1) [0 .. tsize ()] == ((ubyte *) p2) [0 .. tsize ()]; }

}

/* -- delegate ----------------------------------------------------------- */

private void delegate () delegateInit;

class TypeDelegate : TypeInfo
{
override:
    char [] kind () { return "delegate"; }
    int tsize () { return delegateInit.size; }
    byte [] tinit () { return ((byte *) &delegateInit) [0 .. tsize ()]; }
}